#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alloc.h>

typedef unsigned char byte;
typedef unsigned int word;

#include <xref.c>

char *inst3[]={
  "BR",  "BQ",  "BZ",  "BDF", "B1",  "B2",  "B3",  "B4",
  "SKP", "BNQ", "BNZ", "BNF", "BN1", "BN2", "BN3", "BN4"
};
char *inst6[]={
  "IRX", "OUT", "OUT", "OUT", "OUT", "OUT", "OUT", "OUT",
  "_?_", "INP", "INP", "INP", "INP", "INP", "INP", "INP"
};
char *inst7[]={
  "RET", "DIS", "LDXA","STXD","ADC", "SDB", "SHRC","SMB",
  "SAV", "MARK","REQ", "SEQ", "ADCI","SDBI","SHLC","SMBI"
};
char *instc[]={
  "LBR", "LBQ", "LBZ", "LBDF","NOP", "LSNQ","LSNZ","LSNF",
  "LSKP","LBNQ","LBNZ","LBNF","LSIE","LSQ", "LSZ", "LSDF"
};
char *instf[]={
  "LDX", "OR",  "AND", "XOR", "ADD", "SD",  "SHR", "SM",
  "LDI", "ORI", "ANI", "XRI", "ADI", "SDI", "SHL", "SMI"
};

char *inst_reg[]={
  "LDN", "INC", "DEC", NULL,  "LDA", "STR", NULL,  NULL,
  "GLO", "GHI", "PLO", "PHI", NULL,  "SEP", "SEX", NULL
};

char **instr[]={
  NULL,  NULL,  NULL,  inst3, NULL,  NULL,  inst6,  inst7,
  NULL,  NULL,  NULL,  NULL,  instc, NULL,  NULL,  instf
};

byte instr_branch[]={  /* instructions with offset parameter */
  0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
        0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0
};

byte instr_byte[]={    /* instructions with immediate parameter */
  0x7c, 0x7d, 0x7f, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xff, 0
};

byte instr_word[]={    /* instructions with word parameter */
 0xc0, 0xc1, 0xc2, 0xc3, 0xc9, 0xca, 0xcb, 0
};

int call_ret=1;        /* if 1 : print CALL instead of SEP R4 xx xx */

/*
   main disassembly routine
   disassembles instruction from file <f>, current address = <addr>
   stores instruction bytes in <opc>
   disassembled instruction is in <dasm>
   if do_ref==1 : calls addxref to create crossreference table
*/

int disasm(FILE *f, word addr, byte *opc, char *dasm, int do_ref)
{
  byte c;
  byte hi, lo;   /* high/low nyble of instruction in c */
  int ilen=0;      /* instruction length */
  word offset;

  *dasm=0;          /* clear string */
  c=fgetc(f);    hi=(c>>4)&0xf;  lo=c&0xf;
  opc[ilen++]=c;

  if (inst_reg[hi])
  {
    if (c==0)
      sprintf(dasm,"IDL     ");
    else if (c==0xd4 && call_ret)
    {
      opc[ilen++]=fgetc(f);
      opc[ilen++]=fgetc(f);
      offset=(opc[1]<<8)|opc[2];
      sprintf(dasm,"CALL     %04x",offset);
      if (do_ref) addxref(REF_CALL, addr, offset);
    }
    else if (c==0xd5 && call_ret)
      sprintf(dasm,"RETN    ");
    else
      sprintf(dasm,"%-8s R%x", inst_reg[hi], lo);
  }
  else if (hi==6)
  {
    if (lo&7)
      sprintf(dasm,"%-8s D%d", instr[hi][lo], lo&7);
    else
      sprintf(dasm,"%-8s", instr[hi][lo]);
  }
  else if (strchr(instr_branch, c))
  {
    opc[ilen++]=fgetc(f);
    offset=(addr&0xff00)|opc[1];
    sprintf(dasm,"%-8s %04x", instr[hi][lo], offset);
    if (do_ref) addxref(REF_BRANCH, addr, offset);
  }
  else if (strchr(instr_byte, c))
  {
    opc[ilen++]=fgetc(f);
    sprintf(dasm,"%-8s #%02x", instr[hi][lo], opc[1]);
  }
  else if (strchr(instr_word, c))
  {
    opc[ilen++]=fgetc(f);
    opc[ilen++]=fgetc(f);
    offset=(opc[1]<<8)|opc[2];
    sprintf(dasm,"%-8s %04x",instr[hi][lo], offset);
    if (do_ref) addxref(REF_JUMP, addr, offset);
  }
  else
    sprintf(dasm,"%-8s", instr[hi][lo]);
  return ilen;
}

void main(int argc, char **argv)
{
  word addr;
  word prevaddr;
  word start, end;
  int ilen;
  FILE *f;
  byte opc[8];
  char dasm[40];
  int i;
  struct xref far *p;
  struct addrlist far *q;
  int do_hex=1;
  if (argc!=4)
  {
    printf("Usage: 1802 filename start end\n");
    exit(1);
  }
  f=fopen(argv[1],"rb");
  start=strtol(argv[2],0,16);
  end=strtol(argv[3],0,16);

/* generate xrefs */
  fseek(f,(long)start,SEEK_SET);
  addr=start;
  while (addr<end && !feof(f))
    addr+=disasm(f, addr, opc, dasm, 1);

/* output */
  fseek(f, (long)start, SEEK_SET);
  addr=start;
  prevaddr=start;
  p=xref;
  while (addr<end && !feof(f))
  {
    ilen=disasm(f, addr, opc, dasm, 0);
    while (p->ref>=prevaddr && p->ref<=addr)
    {
      printf("%04x  ;",p->ref);
      if (p->ref!=addr)  putchar(';');
      switch(p->type)
      {
        case REF_CALL:    putchar('*');  break;
        case REF_JUMP:    putchar('$');  break;
        case REF_BRANCH:  putchar('#');  break;
        default:          putchar('?');  break;
      }
      for (q=p->list ; q!=NULL ; q=q->n)
        printf("%04x ",q->addr);
      printf("\n");
      p=p->n;
    }
    printf("%04x ", addr);
    if (do_hex)
      for (i=0 ; i<4 ; i++)
        if (i<ilen)
          printf("%02x", opc[i]);
        else
          printf("  ");
    printf("%s\n", dasm);
    prevaddr=addr;
    addr+=ilen;
  }
  printf("\n\n");
  while (p)
  {
    printf("%04x  ;",p->ref);
    switch(p->type)
    {
      case REF_CALL:    putchar('*');  break;
      case REF_JUMP:    putchar('$');  break;
      case REF_BRANCH:  putchar('#');  break;
      default:          putchar('?');  break;
    }
    for (q=p->list ; q!=NULL ; q=q->n)
      printf("%04x ",q->addr);
    printf("\n");
    p=p->n;
  }

  if (missedxrefs)
    printf("%d missed xrefs\n", missedxrefs);
}
