/*
   this deals with the configuration file
    script taal:
    N<binary filename>
    C<code offset> [label]          : define as code
    S<code offset> [n] [label]      : define as subroutine with <n> parameters
    D[t[xx:]]<data offset> [label]  : define as data t=type
         xx = # c/b/w/d per line
     C        characters
     B        bytes, (record length xx)  (default)
     Wlh      words   (low/high)
     Whl      words   (high/low)
     D        double words

     Z        zero terminated strings   (xx = termination character)
     L        length preceded strings

    <n>=PC      0
    <n>=IRQ     1
    <n>=SP      2
    <n>=Params  3
    <n>=CALL    4
    <n>=RET     5

*/


#define DISASSM 0
#define DUMP    1
#define STRINGS 2
#define END     3
#define WORDS   4
#define CH_WORD 5
#define CH_BYTE 6
#define BYTE_CH 7
#define CHARS   8

char datchars[]="cdsewxyzv";

struct fmt {
  int mode;
  int addr;
  struct fmt *n;
} *list;

int addlist(int addr,int mode)
{
  struct fmt *p=list, *q=NULL;
  while (p!=NULL && p->addr<addr)
  {
    q=p;
    p=p->n;
  }
  if (q==NULL)
  {
    q=(struct fmt *)malloc(sizeof(struct fmt));
    if (q==NULL)
    {
      printf("Out of memory reading config. Aborting\n");
      return(-1);
    }
    list=q;
  }
  else
  {
    q->n=(struct fmt *)malloc(sizeof(struct fmt));
    if (q==NULL)
    {
      printf("Out of memory reading config. Aborting\n");
      return(-1);
    }
    q=q->n;
  }
  q->addr=addr;
  q->mode=mode;
  q->n=p;
  return(0);
}

int readlist(char *name)
{
  FILE *f;
  char buf[80], *p, *q;
  int addr;
  f=fopen(name,"r");
  if (f==NULL) return(-1);
  list=NULL;
  while ((p=fgets(buf,80,f))!=NULL)
  {
    buf[strlen(buf)-1]=0;    /* clear newline char */
    for (q=datchars ; *q!=0 && *q!=p[0] ; q++) ;
    if (*q)
    {
      sscanf(p+1,"%x",&addr);
      if (addlist(addr,q-datchars)<0) return(-1);
    }
    else
      switch(p[0])
      {
        case 'f':  /* inputfile */
          inputfile=(char *)malloc(strlen(p));
          if (inputfile==NULL)
          {
            printf("Out of memory allocating string. Aborting\n");
            return(-1);
          }
          strcpy(inputfile,p+1);
          break;
        case 'r':
          if (p[1]=='0')
            sscanf(p+2,"%x",&minxref);
          else
            sscanf(p+2,"%x",&maxxref);
          break;
        case 'b': hexopcodes++; break;
      }
  }
  fclose(f);
  return(0);
}

